
ode23tx = function (F, tspan, y0, rtol = 1e-3, atol = 1e-6) -> (tout, yout)

     /*
        Original Matlab m file by Cleve Moler
        Translated to Shang program
      */

     %   ODE23TX  Solve non-stiff differential equations.  Textbook version of ODE23.
     %
     %   ODE23TX(F,TSPAN,Y0) with TSPAN = [T0 TFINAL] integrates the system
     %   of differential equations y' = f(t,y) from t = T0 to t = TFINAL.  The
     %   initial condition is y(T0) = Y0.  The input argument F is the
     %   name of an M-file, or an inline function, or simply a character string,
     %   defining f(t,y).  This function must have two input arguments, t and y,
     %   and must return a column vector of the derivatives, y'.
     %
     %   With two output arguments, [T,Y] = ODE23TX(...) returns a column 
     %   vector T and an array Y where Y(:,k) is the solution at T(k).
     %
     %   With no output arguments, ODE23TX plots the emerging solution.
     %
     %   ODE23TX(F,TSPAN,Y0,RTOL) uses the relative error tolerance RTOL
     %   instead of the default 1.e-3.
     %
     %   ODE23TX(F,TSPAN,Y0,OPTS) where OPTS = ODESET('reltol',RTOL, ...
     %   'abstol',ATOL,'outputfcn',@PLOTFUN) uses relative error RTOL instead
     %   of 1.e-3, absolute error ATOL instead of 1.e-6, and calls PLOTFUN
     %   instead of ODEPLOT after each successful step.
     %
     %   More than four input arguments, ODE23TX(F,TSPAN,Y0,RTOL,P1,P2,...),
     %   are passed on to F, F(T,Y,P1,P2,...).
     %
     %   ODE23TX uses the Runge-Kutta (2,3) method of Bogacki and Shampine (BS23).
     %
     %   Example    
     %      tspan = [0 2*pi];
     %      y0 = [1 0]';
     %      F = '[0 1; -1 0]*y';
     %      ode23tx(F,tspan,y0);
     %
     %   See also ODE23.
     
     % Initialize variables.
     
     t0 = tspan[1];
     tfinal = tspan[2];
     
     tdir = sign(tfinal - t0);
     
     if tdir == 0
         t = t0;
         y = y0;
         return;
     end
     
     threshold = atol / rtol;
     
     hmax = abs(0.1*(tfinal-t0));
     t = t0;
     y = y0.reshape(y0.length);
     
     % Initialize output.
     
     tstack = stack();
     ystack = stack();
     
     tstack.push(t0);
     ystack.push(y0);
     
     % Compute initial step size.
     
     s1 = F(t, y);
     
     r = norm(s1 / max(abs(y), threshold), inf) + realmin;
     
     h = tdir*0.8*rtol^(1/3)/r;

     % The main loop.
     
     while t != tfinal
       
        hmin = 16*eps*abs(t);
        if abs(h) > hmax
             h = tdir*hmax;
        end

        if abs(h) < hmin
             h = tdir*hmin;
        end
        
        % Stretch the step if t is close to tfinal.
      
        if 1.1*abs(h) >= abs(tfinal - t)
           h = tfinal - t;
        end
        
        % Attempt a step.
      
        s2 = F(t+h/2, y+h/2*s1);
        s3 = F(t+3*h/4, y+3*h/4*s2);
      
        tnew = t + h;
        ynew = y + h*(2*s1 + 3*s2 + 4*s3)/9;
       
        s4 = F(tnew, ynew);
           
        % Estimate the error.
       
        e = h*(-5*s1 + 6*s2 + 8*s3 - 9*s4)/72;
       
        err = norm(e / max(abs(y), abs(ynew), threshold), inf) + realmin;
       
        % Accept the solution if the estimated error is less than the tolerance.
       
        if err <= rtol
           t = tnew;
           y = ynew;
       
           tstack.push(t);
           ystack.push(y);
      
           s1 = s4;     // Reuse final function value to start new step.
        end
        
        % Compute a new step size.
      
        h *= min(5, 0.8*(rtol/err)^(1/3));
      
        % Exit early if step size is too small.
        
        if abs(h) <= hmin
           printf("Step size %e too small at t = %e", h, t);
           t = tfinal;
        end
     end
      
     N = tstack.height;
     
     m = y0.length;
     
     tout = zeros(1, N);
     yout = zeros(m, N);

     for k = N : -1 : 1
          tout[k] = tstack.pop();
          yout[:, k] = ystack.pop();
     end
end
